@@ -2,7 +2,7 @@
  * \file TransverseMercator.cpp
  * \brief Implementation for GeographicLib::TransverseMercator class
  *
- * Copyright (c) Charles Karney (2008, 2009, 2010) <charles@karney.com>
+ * Copyright (c) Charles Karney (2008, 2009, 2010, 2011) <charles@karney.com>
  * and licensed under the LGPL.  For more information, see
  * http://geographiclib.sourceforge.net/
  *
@@ -41,7 +41,7 @@
 
 #include "GeographicLib/TransverseMercator.hpp"
 
-#define GEOGRAPHICLIB_TRANSVERSEMERCATOR_CPP "$Id: TransverseMercator.cpp 6807 2010-02-01 11:26:34Z karney $"
+#define GEOGRAPHICLIB_TRANSVERSEMERCATOR_CPP "$Id: TransverseMercator.cpp 6937 2011-02-01 20:17:13Z karney $"
 
 RCSID_DECL(GEOGRAPHICLIB_TRANSVERSEMERCATOR_CPP)
 RCSID_DECL(GEOGRAPHICLIB_TRANSVERSEMERCATOR_HPP)
@@ -71,6 +71,8 @@ namespace GeographicLib {
   {
     if (!(_a > 0))
       throw GeographicErr("Major radius is not positive");
+    if (!(_f < 1))
+      throw GeographicErr("Minor radius is not positive");
     if (!(_k0 > 0))
       throw GeographicErr("Scale is not positive");
     // If coefficents might overflow an int, convert them to double (and they
@@ -208,17 +210,18 @@ namespace GeographicLib {
   }
 
   const TransverseMercator
-  TransverseMercator::UTM(Constants::WGS84_a(), Constants::WGS84_r(),
-                          Constants::UTM_k0());
+  TransverseMercator::UTM(Constants::WGS84_a<real>(),
+                          Constants::WGS84_r<real>(),
+                          Constants::UTM_k0<real>());
 
   void TransverseMercator::Forward(real lon0, real lat, real lon,
                                    real& x, real& y, real& gamma, real& k)
     const throw() {
     // Avoid losing a bit of accuracy in lon (assuming lon0 is an integer)
     if (lon - lon0 > 180)
-      lon -= lon0 - 360;
-    else if (lon - lon0 <= -180)
       lon -= lon0 + 360;
+    else if (lon - lon0 <= -180)
+      lon -= lon0 - 360;
     else
       lon -= lon0;
     // Now lon in (-180, 180]
@@ -235,8 +238,8 @@ namespace GeographicLib {
       lon = 180 - lon;
     }
     real
-      phi = lat * Constants::degree(),
-      lam = lon * Constants::degree();
+      phi = lat * Math::degree<real>(),
+      lam = lon * Math::degree<real>();
     // phi = latitude
     // phi' = conformal latitude
     // psi = isometric latitude
@@ -255,13 +258,13 @@ namespace GeographicLib {
     //   cosh(etap) = 1/denom                  = 1/denom
     //   sinh(etap) = cos(phi')*sin(lam)/denom = sech(psi)*sin(lam)/denom
     real etap, xip;
-    if (lat < 90) {
+    if (lat != 90) {
       real
         c = max(real(0), cos(lam)), // cos(pi/2) might be negative
-        tau = tanx(phi),
+        tau = tan(phi),
         secphi = Math::hypot(real(1), tau),
-        sig = sinh( eatanhe(sin(phi)) ),
-        taup = (Math::hypot(real(1), sig) * tau - sig * secphi);
+        sig = sinh( eatanhe(tau / secphi) ),
+        taup = Math::hypot(real(1), sig) * tau - sig * secphi;
       xip = atan2(taup, c);
       // Used to be
       //   etap = Math::atanh(sin(lam) / cosh(psi));
@@ -280,7 +283,7 @@ namespace GeographicLib {
       // instead of psip).
       k = sqrt(_e2m + _e2 * sq(cos(phi))) * secphi / Math::hypot(taup, c);
     } else {
-      xip = Constants::pi()/2;
+      xip = Math::pi<real>()/2;
       etap = 0;
       gamma = lam;
       k = _c;
@@ -372,8 +375,8 @@ namespace GeographicLib {
     // Gauss-Krueger TM.
     gamma -= atan2(yi1, yr1);
     k *= _b1 * Math::hypot(yr1, yi1);
-    gamma /= Constants::degree();
-    y = _a1 * _k0 * (backside ? Constants::pi() - xi : xi) * latsign;
+    gamma /= Math::degree<real>();
+    y = _a1 * _k0 * (backside ? Math::pi<real>() - xi : xi) * latsign;
     x = _a1 * _k0 * eta * lonsign;
     if (backside)
       gamma = 180 - gamma;
@@ -386,7 +389,7 @@ namespace GeographicLib {
     const throw() {
     // This undoes the steps in Forward.  The wrinkles are: (1) Use of the
     // reverted series to express zeta' in terms of zeta. (2) Newton's method
-    // to solve for phi in terms of psi.
+    // to solve for phi in terms of tan(phi).
     real
       xi = y / (_a1 * _k0),
       eta = x / (_a1 * _k0);
@@ -396,9 +399,9 @@ namespace GeographicLib {
       etasign = eta < 0 ? -1 : 1;
     xi *= xisign;
     eta *= etasign;
-    bool backside = xi > Constants::pi()/2;
+    bool backside = xi > Math::pi<real>()/2;
     if (backside)
-      xi = Constants::pi() - xi;
+      xi = Math::pi<real>() - xi;
     real
       c0 = cos(2 * xi), ch0 = cosh(2 * eta),
       s0 = sin(2 * xi), sh0 = sinh(2 * eta),
@@ -442,22 +445,23 @@ namespace GeographicLib {
       s = sinh(etap),
       c = max(real(0), cos(xip)), // cos(pi/2) might be negative
       r = Math::hypot(s, c);
-    if (r > 0) {
+    if (r != 0) {
       lam = atan2(s, c);        // Krueger p 17 (25)
-      // Use Newton's< method to solve for tau
+      // Use Newton's method to solve for tau
       real
         taup = sin(xip)/r,
-        tau = taup;
+        tau = taup,
+        stol = tol * max(real(1), abs(taup));
       // min iterations = 1, max iterations = 2; mean = 1.99
       for (int i = 0; i < numit; ++i) {
         real
           tau1 = Math::hypot(real(1), tau),
           sig = sinh( eatanhe( tau / tau1 ) ),
-          sig1 =  Math::hypot(real(1), sig),
-          dtau = - (sig1 * tau - sig * tau1 - taup) * (1 + _e2m * sq(tau)) /
-          ( (sig1 * tau1 - sig * tau) * _e2m * tau1 );
+          taupa = Math::hypot(real(1), sig) * tau - sig * tau1,
+          dtau = (taup - taupa) * (1 + _e2m * sq(tau)) /
+          ( _e2m * tau1 * Math::hypot(real(1), taupa) );
         tau += dtau;
-        if (abs(dtau) < tol * max(real(1), tau))
+        if (!(abs(dtau) >= stol))
           break;
       }
       phi = atan(tau);
@@ -465,12 +469,12 @@ namespace GeographicLib {
       // Note cos(phi') * cosh(eta') = r
       k *= sqrt(_e2m + _e2 * sq(cos(phi))) * Math::hypot(real(1), tau) * r;
     } else {
-      phi = Constants::pi()/2;
+      phi = Math::pi<real>()/2;
       lam = 0;
       k *= _c;
     }
-    lat = phi / Constants::degree() * xisign;
-    lon = lam / Constants::degree();
+    lat = phi / Math::degree<real>() * xisign;
+    lon = lam / Math::degree<real>();
     if (backside)
       lon = 180 - lon;
     lon *= etasign;
@@ -481,7 +485,7 @@ namespace GeographicLib {
       lon += lon0 + 360;
     else
       lon += lon0;
-    gamma /= Constants::degree();
+    gamma /= Math::degree<real>();
     if (backside)
       gamma = 180 - gamma;
     gamma *= xisign * etasign;
