@@ -2,7 +2,7 @@
  * \file TransverseMercatorExact.cpp
  * \brief Implementation for GeographicLib::TransverseMercatorExact class
  *
- * Copyright (c) Charles Karney (2008, 2009, 2010) <charles@karney.com>
+ * Copyright (c) Charles Karney (2008, 2009, 2010, 2011) <charles@karney.com>
  * and licensed under the LGPL.  For more information, see
  * http://geographiclib.sourceforge.net/
  *
@@ -29,6 +29,8 @@
  * <tr><td>k^2    <td>mu      <td>elliptic function parameter
  * <tr><td>k'^2   <td>mv      <td>elliptic function complementary parameter
  * <tr><td>m      <td>k       <td>scale
+ * <tr><td>zeta   <td>zeta    <td>complex longitude = Mercator = chi in paper
+ * <tr><td>s      <td>sigma   <td>complex GK = zeta in paper
  * </table></center>
  *
  * Minor alterations have been made in some of Lee's expressions in an
@@ -39,7 +41,7 @@
 
 #include "GeographicLib/TransverseMercatorExact.hpp"
 
-#define GEOGRAPHICLIB_TRANSVERSEMERCATOREXACT_CPP "$Id: TransverseMercatorExact.cpp 6813 2010-02-03 13:57:19Z karney $"
+#define GEOGRAPHICLIB_TRANSVERSEMERCATOREXACT_CPP "$Id: TransverseMercatorExact.cpp 6937 2011-02-01 20:17:13Z karney $"
 
 RCSID_DECL(GEOGRAPHICLIB_TRANSVERSEMERCATOREXACT_CPP)
 RCSID_DECL(GEOGRAPHICLIB_TRANSVERSEMERCATOREXACT_HPP)
@@ -74,34 +76,38 @@ namespace GeographicLib {
       throw GeographicErr("Major radius is not positive");
     if (!(_r > 0))
       throw GeographicErr("Inverse flattening is not positive");
+    if (!(_f < 1))
+      throw GeographicErr("Minor radius is not positive");
     if (!(_k0 > 0))
       throw GeographicErr("Scale is not positive");
   }
 
   const TransverseMercatorExact
-  TransverseMercatorExact::UTM(Constants::WGS84_a(), Constants::WGS84_r(),
-                               Constants::UTM_k0());
+  TransverseMercatorExact::UTM(Constants::WGS84_a<real>(),
+                               Constants::WGS84_r<real>(),
+                               Constants::UTM_k0<real>());
 
   // tau = tan(phi), taup = sinh(psi)
   Math::real TransverseMercatorExact::taup(real tau) const throw() {
     real
       tau1 = Math::hypot(real(1), tau),
-      sig = sinh( _e * Math::atanh(_e * tau / tau1) ),
-      sig1 = Math::hypot(real(1), sig);
-    return tau * sig1 - sig * tau1;
+      sig = sinh( _e * Math::atanh(_e * tau / tau1) );
+    return Math::hypot(real(1), sig) * tau - sig * tau1;
   }
 
   Math::real TransverseMercatorExact::taupinv(real taup) const throw() {
-    real tau = taup;
+    real
+      tau = taup,
+      stol = tol * max(real(1), abs(taup));
     for (int i = 0; i < numit; ++i) {
       real
         tau1 = Math::hypot(real(1), tau),
         sig = sinh( _e * Math::atanh(_e * tau / tau1 ) ),
-        sig1 =  Math::hypot(real(1), sig),
-        dtau = - (sig1 * tau - sig * tau1 - taup) * (1 + _mv * sq(tau)) /
-        ( (sig1 * tau1 - sig * tau) * _mv * tau1 );
+        taupa = Math::hypot(real(1), sig) * tau - sig * tau1,
+        dtau = (taup - taupa) * (1 + _mv * sq(tau)) /
+        ( _mv * tau1 * Math::hypot(real(1), taupa) );
       tau += dtau;
-      if (abs(dtau) < tol * max(real(1), tau))
+      if (!(abs(dtau) >= stol))
         break;
     }
     return tau;
@@ -142,9 +148,9 @@ namespace GeographicLib {
   bool TransverseMercatorExact::zetainv0(real psi, real lam, real& u, real& v)
     const throw() {
     bool retval = false;
-    if (psi < -_e * Constants::pi()/4 &&
-        lam > (1 - 2 * _e) * Constants::pi()/2 &&
-        psi < lam - (1 - _e) * Constants::pi()/2) {
+    if (psi < -_e * Math::pi<real>()/4 &&
+        lam > (1 - 2 * _e) * Math::pi<real>()/2 &&
+        psi < lam - (1 - _e) * Math::pi<real>()/2) {
       // N.B. this branch is normally not taken because psi < 0 is converted
       // psi > 0 by Forward.
       //
@@ -156,14 +162,14 @@ namespace GeographicLib {
       // Inverting this gives:
       real
         psix = 1 - psi / _e,
-        lamx = (Constants::pi()/2 - lam) / _e;
+        lamx = (Math::pi<real>()/2 - lam) / _e;
       u = Math::asinh(sin(lamx) / Math::hypot(cos(lamx), sinh(psix))) *
         (1 + _mu/2);
       v = atan2(cos(lamx), sinh(psix)) * (1 + _mu/2);
       u = _Eu.K() - u;
       v = _Ev.K() - v;
-    } else if (psi < _e * Constants::pi()/2 &&
-               lam > (1 - 2 * _e) * Constants::pi()/2) {
+    } else if (psi < _e * Math::pi<real>()/2 &&
+               lam > (1 - 2 * _e) * Math::pi<real>()/2) {
       // At w = w0 = i * Ev.K(), we have
       //
       //     zeta = zeta0 = i * (1 - _e) * pi/2
@@ -176,14 +182,14 @@ namespace GeographicLib {
       // When inverting this, we map arg(w - w0) = [-90, 0] to
       // arg(zeta - zeta0) = [-90, 180]
       real
-        dlam = lam - (1 - _e) * Constants::pi()/2,
+        dlam = lam - (1 - _e) * Math::pi<real>()/2,
         rad = Math::hypot(psi, dlam),
         // atan2(dlam-psi, psi+dlam) + 45d gives arg(zeta - zeta0) in range
         // [-135, 225).  Subtracting 180 (since multiplier is negative) makes
         // range [-315, 45).  Multiplying by 1/3 (for cube root) gives range
         // [-105, 15).  In particular the range [-90, 180] in zeta space maps
         // to [-90, 0] in w space as required.
-        ang = atan2(dlam-psi, psi+dlam) - real(0.75) * Constants::pi();
+        ang = atan2(dlam-psi, psi+dlam) - real(0.75) * Math::pi<real>();
       // Error using this guess is about 0.21 * (rad/e)^(5/3)
       retval = rad < _e * taytol;
       rad = Math::cbrt(3 / (_mv * _e) * rad);
@@ -197,8 +203,8 @@ namespace GeographicLib {
       v = Math::asinh(sin(lam) / Math::hypot(cos(lam), sinh(psi)));
       u = atan2(sinh(psi), cos(lam));
       // But scale to put 90,0 on the right place
-      u *= _Eu.K() / (Constants::pi()/2);
-      v *= _Eu.K() / (Constants::pi()/2);
+      u *= _Eu.K() / (Math::pi<real>()/2);
+      v *= _Eu.K() / (Math::pi<real>()/2);
     }
     return retval;
   }
@@ -231,7 +237,7 @@ namespace GeographicLib {
       if (trip)
         break;
       real delw2 = sq(delu) + sq(delv);
-      if (delw2 < stol2)
+      if (!(delw2 >= stol2))
         ++trip;
     }
   }
@@ -294,7 +300,7 @@ namespace GeographicLib {
         rad = Math::hypot(xi, deta),
         // Map the range [-90, 180] in sigma space to [-90, 0] in w space.  See
         // discussion in zetainv0 on the cut for ang.
-        ang = atan2(deta-xi, xi+deta) - real(0.75) * Constants::pi();
+        ang = atan2(deta-xi, xi+deta) - real(0.75) * Math::pi<real>();
       // Error using this guess is about 0.068 * rad^(5/3)
       retval = rad < 2 * taytol;
       rad = Math::cbrt(3 / _mv * rad);
@@ -332,7 +338,7 @@ namespace GeographicLib {
       if (trip)
         break;
       real delw2 = sq(delu) + sq(delv);
-      if (delw2 < tol2)
+      if (!(delw2 >= tol2))
         ++trip;
     }
   }
@@ -368,9 +374,9 @@ namespace GeographicLib {
     const throw() {
     // Avoid losing a bit of accuracy in lon (assuming lon0 is an integer)
     if (lon - lon0 > 180)
-      lon -= lon0 - 360;
-    else if (lon - lon0 <= -180)
       lon -= lon0 + 360;
+    else if (lon - lon0 <= -180)
+      lon -= lon0 - 360;
     else
       lon -= lon0;
     // Now lon in (-180, 180]
@@ -387,8 +393,8 @@ namespace GeographicLib {
       lon = 180 - lon;
     }
     real
-      phi = lat * Constants::degree(),
-      lam = lon * Constants::degree(),
+      phi = lat * Math::degree<real>(),
+      lam = lon * Math::degree<real>(),
       tau = tanx(phi);
 
     // u,v = coordinates for the Thompson TM, Lee 54
@@ -417,7 +423,7 @@ namespace GeographicLib {
     zeta(u, snu, cnu, dnu, v, snv, cnv, dnv, tau, lam);
     tau=taupinv(tau);
     Scale(tau, lam, snu, cnu, dnu, snv, cnv, dnv, gamma, k);
-    gamma /= Constants::degree();
+    gamma /= Math::degree<real>();
     if (backside)
       gamma = 180 - gamma;
     gamma *= latsign * lonsign;
@@ -458,16 +464,16 @@ namespace GeographicLib {
       zeta(u, snu, cnu, dnu, v, snv, cnv, dnv, tau, lam);
       tau = taupinv(tau);
       phi = atan(tau);
-      lat = phi / Constants::degree();
-      lon = lam / Constants::degree();
+      lat = phi / Math::degree<real>();
+      lon = lam / Math::degree<real>();
     } else {
       tau = overflow;
-      phi = Constants::pi()/2;
+      phi = Math::pi<real>()/2;
       lat = 90;
       lon = lam = 0;
     }
     Scale(tau, lam, snu, cnu, dnu, snv, cnv, dnv, gamma, k);
-    gamma /= Constants::degree();
+    gamma /= Math::degree<real>();
     if (backside)
       lon = 180 - lon;
     lon *= lonsign;
